/**
 *    FrameCreator.java
 *    Copyright (C) 2010 New Zealand Digital Library, http://expeditee.org
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.expeditee.gui;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.LinkedList;
import java.util.List;

import org.expeditee.agents.ExistingFramesetException;
import org.expeditee.agents.InvalidFramesetNameException;
import org.expeditee.core.Colour;
import org.expeditee.items.Item;
import org.expeditee.items.Text;

public class FrameCreator {

	public static final int INDENT_FROM_TITLE = 20;

	private int START_Y;

	private int START_X;

	private Frame _current = null;

	private int _lastX;

	private int _maxX = 0;

	private int _lastY;

	// Master copy of next & previous links
	private Item _Mnext;

	private Item _Mprev;

	private Item _Mfirst;

	private Frame _firstFrame;

	private boolean _multiColumn;

	public enum ExistingFramesetOptions {
		/**
		 * Do not attempt to use any existing frames in specified framesets.
		 */
		AppendSegregatedFrames,
		/**
		 * Override any existing frames in specified framesets.
		 */
		OverrideExistingFrames,
		/**
		 * Use the existing frames in the framesets to reobtain what the state of the
		 * FrameCreator should be.
		 */
		AppendAfterLastItem
	}

	private final List<Frame> framesCreated = new LinkedList<Frame>();

	public FrameCreator(String frameTitle) {
		this(DisplayController.getCurrentFrame().getFramesetName(), DisplayController.getCurrentFrame().getPath(),
				frameTitle, ExistingFramesetOptions.AppendSegregatedFrames, false);
	}

	public FrameCreator(String framesetName, String path, String frameTitle, ExistingFramesetOptions establishState, boolean multiColumn) {
		switch (establishState) {
			case OverrideExistingFrames:
				try {
					FrameIO.SuspendCache();
					Path framesetDir = Files.createTempDirectory("temporaryExpediteeFrameset");
					initialise(framesetName, framesetDir.toAbsolutePath(), frameTitle, multiColumn);
					FrameIO.EnableCache();
					FrameIO.deleteFrameset(framesetName);
					FrameIO.moveFrameset(framesetName, framesetDir.toAbsolutePath().toString() + File.separator, path);
					_current.setPath(path);
				} catch (IOException e) {
					e.printStackTrace();
				}
				break;
			case AppendSegregatedFrames:
				initialise(framesetName, Paths.get(path), frameTitle, multiColumn);
				break;
			case AppendAfterLastItem:
				initialiseReobtainState(framesetName, Paths.get(path), frameTitle, multiColumn);
				break;
		}
	}

	private void initialiseReobtainState(String framesetName, Path path, String frameTitle, boolean multiColumn) {
		_multiColumn = multiColumn;
		_Mnext = createButton("@Next", null, null, 10, 15);
		_Mprev = createButton("@Previous", null, null, _Mnext.getBoundsWidth() + _Mnext.getAnchorRight() + 20, 15);
		_Mfirst = createButton("@First", null, null, _Mprev.getBoundsWidth() + _Mprev.getAnchorRight() + 20, 15);

		int lastNumber = FrameIO.getLastNumber(framesetName);
		for (int i = 1; i <= lastNumber; i++) {
			Frame frame = FrameIO.LoadFrame(framesetName + i, path.toAbsolutePath().toString() + File.separator);
			this.framesCreated.add(frame);
		}
		if (this.framesCreated.isEmpty()) {
			initialise(framesetName, path, frameTitle, multiColumn);
			return;
		}
		_firstFrame = this.framesCreated.get(0);
		_current = this.framesCreated.get(this.framesCreated.size() - 1);

		createNextFrame();
		
		Paths.get(_current.getFramePathLogical());
	}
	
	private void initialise(String framesetName, Path path, String frameTitle, boolean multiColumn) {
		_multiColumn = multiColumn;
		_Mnext = createButton("@Next", null, null, 10, 15);
		_Mprev = createButton("@Previous", null, null, _Mnext.getBoundsWidth() + _Mnext.getAnchorRight() + 20, 15);
		_Mfirst = createButton("@First", null, null, _Mprev.getBoundsWidth() + _Mprev.getAnchorRight() + 20, 15);
			
		try {
			Frame toUse;
			if (FrameIO.canAccessFrameset(framesetName, path)) {
				toUse = FrameIO.CreateFrame(framesetName, frameTitle, null);
			} else {
				toUse = FrameIO.CreateFrameset(framesetName, path.toAbsolutePath().toString() + File.separator, true);
			}
		
			this.framesCreated.add(toUse);
			resetGlobals(toUse);
			_firstFrame = toUse;
		} catch (InvalidFramesetNameException e) {
			e.printStackTrace();
		} catch (ExistingFramesetException e) {
			// This should never happen as we are allowing for an override of existing frameset.
		}
	}

	/**
	 * Creates a text item that looks like a clickable button.
	 * 
	 * @param text
	 *            the caption for the button
	 * @param x
	 *            the x position for the button. Null if the button is anchored to
	 *            the right of the screen.
	 * @param y
	 *            the y position for the button. Null if the button is anchored to
	 *            the bottom of the screen.
	 * @param right
	 *            the distance the button is anchored from the right of this screen.
	 *            Null if an absolute position is used.
	 * @param bottom
	 *            the distance the button is to be anchored from the bottom of the
	 *            screen. Null if the button is given an absolute position.
	 * @return the newly created button.
	 */
	public static Item createButton(String text, Integer x, Integer y, Integer right, Integer bottom) {
		Text button = new Text(text);

		button.setBackgroundColor(Colour.LIGHT_GREY);
		button.setBorderColor(Colour.DARK_GREY);
		button.setThickness(2.0F);
		if (bottom != null)
			button.setAnchorBottom(bottom);
		if (x != null)
			button.setX(x);
		if (right != null)
			button.setAnchorRight(right);
		if (y != null)
			button.setY(y);

		button.invalidateBounds();

		return button;
	}

	public String getName() {
		return _firstFrame.getName();
	}

	/**
	 * Creates the next frame in the frameset, with a previous button already added
	 * and linked to the last frame. _current then gets updated to point at the
	 * newly created Frame, and _lastY is reset
	 */
	public boolean createNextFrame() {
		try {
			Frame newFrame = FrameIO.CreateFrame(_current.getFramesetName(), _current.getTitle(), null);

			this.framesCreated.add(newFrame);

			// add link to previous frame
			// _prev =
			addPreviousButton(newFrame, _current.getName());

			// add link to new frame
			// _next =
			addNextButton(_current, newFrame.getName());

			// add link to new frame
			addFirstButton(newFrame, _firstFrame.getName());

			FrameIO.SaveFrame(_current, false);

			resetGlobals(newFrame);
			_maxX = 0;
			return true;
		} catch (Exception e) {
			return false;
		}
	}

	public List<Frame> getFramesCreated() {
		return this.framesCreated;
	}

	private void resetGlobals(Frame toUse) {
		Text title = toUse.getTitleItem();
		START_X = INDENT_FROM_TITLE + title.getX();
		START_Y = getYStart(title);
		_lastY = START_Y;
		_lastX = START_X;
		// Check for @Start
		for (Item it : toUse.getItems()) {
			if (it instanceof Text) {
				Text t = (Text) it;
				if (t.getText().toLowerCase().equals("@start") || t.getText().toLowerCase().startsWith("@start:")) {
					t.stripFirstWord();

					if (t.getText().equals("")) {
						_lastY = t.getY();
						_lastX = t.getX();
						t.delete();
						break;
					}
				}
			}
		}
		_current = toUse;
	}

	public boolean addItem(Item toAdd, boolean bSave) {
		try {
			// if we have reached the end of the Y axis, try moving over on the
			// X axis
			if (_lastY >= _Mprev.getY() - _Mprev.getBoundsHeight()) {
				_lastX = _maxX + 10;
				_lastY = START_Y;

				// if there is no more room on the X axis, we have to start a
				// new frame
				if (!_multiColumn || toAdd.getBoundsWidth() + _lastX > DisplayController.getFramePaintAreaWidth()) {
					// Make sure text items that are created on the current
					// frame are removed
					_current.removeItem(toAdd);
					createNextFrame();
				}
			}

			toAdd.setPosition(_lastX, _lastY + toAdd.getBoundsHeight() / 2);
			toAdd.setOffset(0, 0);
			toAdd.setID(_current.getNextItemID());
			toAdd.setRightMargin(DisplayController.getFramePaintAreaWidth(), true);

			_current.addItem(toAdd);
			// _current.addAllItems(items);
			if (bSave)
				save();

			_lastY = toAdd.getY() + toAdd.getBoundsHeight() / 2;
			_maxX = Math.max(toAdd.getX() + toAdd.getBoundsWidth(), _maxX);

			return true;
		} catch (Exception e) {
			return false;
		}
	}

	public Text addText(String toAdd, Colour c, String link, String action, boolean bSave) {
		Text text = _current.createNewText(toAdd);
		if (c != null)
			text.setColor(c);
		text.setLink(link);
		text.setAction(action);

		addItem(text, bSave);

		return text;
	}

	public void save() {
		FrameIO.ForceSaveFrame(_current);
	}

	public int getLastY() {
		return _lastY;
	}

	public void setLastY(int lastY) {
		_lastY = lastY;
	}

	public Frame getCurrentFrame() {
		return _current;
	}

	/**
	 * Returns the Frame name of the current frame for the FrameCreator
	 * 
	 * @return The current frame for the FrameCreator
	 */
	public String getCurrent() {
		if (_current == null)
			return null;

		return _current.getName();
	}

	public Frame getFirstFrame() {
		return _firstFrame;
	}

	public Item addNextButton(Frame current, String link) {
		return addButton(_Mnext, current, link);
	}

	public Item addPreviousButton(Frame current, String link) {
		return addButton(_Mprev, current, link);
	}

	public Item addFirstButton(Frame current, String link) {
		return addButton(_Mfirst, current, link);
	}

	public Item addButton(Item template, Frame current, String link) {
		// add link to new frame
		Item previousButton = template.copy();
		previousButton.setID(current.getNextItemID());
		previousButton.setLink(link);
		previousButton.setLinkHistory(false);
		previousButton.setLinkMark(false);
		// previousButton.setPermission(new
		// PermissionPair(UserAppliedPermission.followLinks));
		current.addItem(previousButton);

		return previousButton;
	}

	public void addSpace(int space) {
		_lastY += space;
	}

	public static int getYStart(Item title) {
		return title.getY() + title.getBoundsHeight();
	}

	public void setTitle(String titleText) {
		_current.setTitle(titleText);
	}
}
